/* xoverkill - exploding xkill variant
 *
 * Copyright 2005 Michael J M Thomson
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * THE AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
 * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include <X11/Xlib.h>
#include <X11/cursorfont.h>
#include <X11/Xmu/WinUtil.h>
#include <X11/extensions/shape.h>
#include <X11/Xatom.h>
#include <stdbool.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <sys/time.h>
#include <time.h>
#include <unistd.h>
#ifdef SOUND_ENABLED
#include <fcntl.h>
#endif

Display* display;
int screen_number;
Screen* screen;
int shape_event, shape_error;

enum { KILL_VICTIM, ICONIFY_VICTIM } victim_op;
Window victim_window;
int victim_x, victim_y;
unsigned victim_width, victim_height;

int selected_x, selected_y;
int h_overlap, v_overlap;

int overlay_x, overlay_y;
unsigned overlay_width, overlay_height;
Window overlay_window;
Pixmap overlay_buffer;

#define FRAMES_PER_SECOND 10

struct FragArea {
  int x, y;
  unsigned width, height;
} frag_area;
int frag_quadrant = 1;
int frag_rate;
bool disposed_of_victim = false;

typedef struct Frag Frag;
typedef enum FragStage {
  FRAG_EXPANDING,
  FRAG_SMOKING,
  FRAG_CLEARING,
  FRAG_CLEARED
} FragStage;

#define MIN_RADIUS 2
struct Frag {
  int x, y;
  unsigned current_radius, max_radius;
  FragStage stage;
  Frag* next;
};
Frag* frag_list = NULL;
unsigned n_active_frags = 0, max_active_frags = 0;
bool no_more_frags = false;

#ifdef SOUND_ENABLED
const char* sound_device = "/dev/dsp";
int sound_fd;
bool sound_ok = false;
#define SOUND_PULSE_BEGIN 8192 / FRAMES_PER_SECOND
int sound_level, sound_pulse;
#endif

Pixmap stipple_pixmap = 0;
GC stipple_gcs[2];
#define GRADIENT_SIZE 128
GC gradient_gcs[GRADIENT_SIZE];
GC smoke_gc;
Pixmap shape_pixmap;
enum ShapeGCType {
  SHAPE_GC_CLEAR_STIPPLED,
  SHAPE_GC_CLEAR_SOLID,
  SHAPE_GC_SET_STIPPLED,
  N_SHAPE_GC_TYPES
};
GC shape_gcs[N_SHAPE_GC_TYPES];

int randn(int max) {
  return (int)((max + 1.0) * (rand() / (double)(RAND_MAX)));
}

int min(int a, int b) {
  return (a < b ? a : b);
}

int max(int a, int b) {
  return (a > b ? a : b);
}

int clip(int v, int min, int max) {
  return (v > min ? (v < max ? v : max) : min);
}

Window select_window(int* selected_x, int* selected_y);
Window find_outermost_window(Window inner);

void overlay_create();
void overlay_show();
void overlay_blit_area(int x, int y, unsigned width, unsigned height);

Pixmap take_screenshot(int x, int y, unsigned width, unsigned height);

void kill_or_iconify_victim();

void animation_init();
void animation_run();
bool animation_advance();

#ifdef SOUND_ENABLED
void sound_init();
void sound_generate_frame();
#endif

Frag* frag_create();
void frag_advance(Frag*);
void frag_destroy(Frag*);

int main(int argc, char** argv) {
  const char* usage = \
    "Usage: xoverkill [options]\n"
    "Options may be any of:\n"
    "  -display DISPLAYNAME    X server to connect to\n"
    "  -id WINDOWID            Window to kill process of or iconify\n"
    "  -iconify                Iconify window instead of killing process\n"
#ifdef SOUND_ENABLED
    "  -sounddevice DEVNAME    Sound device to use instead of /dev/dsp\n"
    "  -nosound                Disable sound effect\n"
    ;
    bool nosound = false
#endif
  ;
  const char* display_name = 0;
  int i;
  Window unused1;
  unsigned int unused2;

  victim_window = None;
  victim_op = KILL_VICTIM;
  for (i = 1; i < argc; ++i) {
    const char* option = argv[i];
    const char* value = (i + 1 < argc ? argv[i + 1] : "");

    if (strcmp(option, "-display") == 0 && ++i < argc)
      display_name = value;
    else if (strcmp(option, "-id") == 0 && ++i < argc) {
      char* endp;
      victim_window = strtol(value, &endp, 0);
      if (endp == value) {
        printf("Invalid value '%s' for option '-id'\n", value);
        printf("%s\n", usage);
        return -1;
      }
    }
    else if (strcmp(option, "-iconify") == 0) {
      victim_op = ICONIFY_VICTIM;
    }
#ifdef SOUND_ENABLED
    else if (strcmp(option, "-sounddevice") == 0 && ++i < argc) {
      sound_device = argv[i];
    }
    else if (strcmp(option, "-nosound") == 0) {
      nosound = true;
    }
#endif
    else {
      printf("Unrecognized option '%s'\n", option);
      printf("%s\n", usage);
      return -1;
    }
  }

  display = XOpenDisplay(display_name);
  if (!display) {
    printf("Unable to open display (%s)\n",
           display_name ? display_name
                        : "default");
    return -1;
  }
  screen_number = DefaultScreen(display);
  screen = ScreenOfDisplay(display, screen_number);
  if (!XShapeQueryExtension(display, &shape_event, &shape_error)) {
    printf("Required X Shape Extension not available on X display.\n");
    return -1;
  }

  srand(time(NULL));

  if (victim_window == None) {
    printf("Click on window %s, or on background to cancel...\n",
           victim_op == KILL_VICTIM ? "whose process you wish to kill"
                                    : "you wish to iconify");
    victim_window = select_window(&selected_x, &selected_y);
    if (!victim_window) {
      printf("Cancelled.\n");
      return 0;
    }
  }
  else {
    victim_window = find_outermost_window(victim_window);
    selected_x = selected_y = -1;
  }
  XGetGeometry(display, victim_window, &unused1,
               &victim_x, &victim_y, &victim_width, &victim_height, &unused2, &unused2);
  if (selected_x != -1 && selected_y != -1) {
    selected_x -= victim_x;
    selected_y -= victim_y;
  }
  else {
    selected_x = victim_width / 2;
    selected_y = victim_height / 2;
  }

  overlay_create();
  overlay_show();

  animation_init();
#ifdef SOUND_ENABLED
  if (!nosound)
    sound_init();
#endif

  animation_run();

  return 0;
}

Window select_window(int* selected_x, int* selected_y) {
  Cursor cursor;
  Window window;

  cursor = XCreateFontCursor(display, XC_crosshair);
  XGrabPointer(display, screen->root, True,
               ButtonReleaseMask, GrabModeAsync, GrabModeAsync,
               screen->root, cursor, CurrentTime);

  window = 0;
  while (!window) {
    XEvent xevent;
    XNextEvent(display, &xevent);
    switch (xevent.type) {
      case ButtonRelease:
        if (xevent.xbutton.subwindow == None)
          return None;
        window = find_outermost_window(xevent.xbutton.subwindow);
        *selected_x = xevent.xbutton.x_root;
        *selected_y = xevent.xbutton.y_root;
        break;
    }
  }

  XUngrabPointer(display, CurrentTime);
  XFreeCursor(display, cursor);

  return window;
}

Window find_outermost_window(Window inner) {
  Window root, parent, *children = NULL;
  unsigned n_children;

  XQueryTree(display, inner,
             &root, &parent, &children, &n_children);
  if (children)
    XFree(children);
  if (parent == screen->root)
    return inner;
  else
    return find_outermost_window(parent);
}

void overlay_create() {
  XSetWindowAttributes cw_values;

  h_overlap = victim_width / 8;
  v_overlap = victim_height / 8;

  overlay_x = victim_x - h_overlap;
  overlay_y = victim_y - v_overlap;
  overlay_width = victim_width + h_overlap * 2;
  overlay_height = victim_height + v_overlap * 2;

  cw_values.override_redirect = True;
  cw_values.colormap = screen->cmap;
  cw_values.event_mask = None;
  overlay_window = XCreateWindow(display, screen->root,
                                 overlay_x, overlay_y, overlay_width, overlay_height, 0,
                                 screen->root_depth, InputOutput, CopyFromParent,
                                 CWOverrideRedirect |
                                 CWColormap |
                                 CWEventMask,
                                 &cw_values);
  overlay_buffer = XCreatePixmap(display, overlay_window,
                                 overlay_width, overlay_height,
                                 screen->root_depth);
  XSetWindowBackgroundPixmap(display, overlay_window, overlay_buffer);
}

void overlay_show() {
  Pixmap screenshot;
  XGCValues gc_values;
  GC gc;

  screenshot = take_screenshot(overlay_x, overlay_y, overlay_width, overlay_height);
  gc_values.function = GXcopy;
  gc = XCreateGC(display, overlay_buffer,
                 GCFunction, &gc_values);
  XCopyArea(display, screenshot, overlay_buffer, gc,
            0, 0, overlay_width, overlay_height, 0, 0);
  XFreePixmap(display, screenshot);
  XFreeGC(display, gc);

  XMapRaised(display, overlay_window);
}

void overlay_blit_area(int x, int y, unsigned width, unsigned height) {
  XClearArea(display, overlay_window, x, y, width, height, False);
}

Pixmap take_screenshot(int x, int y, unsigned width, unsigned height) {
  Pixmap screenshot;
  XGCValues gc_values;
  GC gc;

  screenshot = XCreatePixmap(display, screen->root,
                             overlay_width, overlay_height,
                             screen->root_depth);

  gc_values.function = GXcopy;
  gc_values.subwindow_mode = IncludeInferiors;
  gc = XCreateGC(display, screen->root,
                 GCFunction |
                 GCSubwindowMode,
                 &gc_values);

  XCopyArea(display, screen->root, screenshot, gc,
            x, y, width, height, 0, 0);

  XFreeGC(display, gc);

  return screenshot;
}

void kill_or_iconify_victim() {
  Window client_window;

  client_window = XmuClientWindow(display, victim_window);

  if (victim_op == KILL_VICTIM) {
    Atom atom_NET_WM_PID;
    pid_t pid = 0;

    printf("xoverkill: Killing client process of window %p...\n", (void*)client_window);

    /* Try _NET_WM_PID value (provided WM_CLIENT_MACHINE matches own host) */
    atom_NET_WM_PID = XInternAtom(display, "_NET_WM_PID", True);
    if (atom_NET_WM_PID) {
      Atom actual_type;
      int actual_format;
      unsigned long nitems, bytes_after;
      pid_t* pid_data = 0;
      char* string_data = 0;
      char hostname[256] = { 0, };

      if (XGetWindowProperty(display, client_window, atom_NET_WM_PID,
                             0, 1, False, XA_CARDINAL,
                             &actual_type, &actual_format, &nitems, &bytes_after,
                             (unsigned char**)&pid_data) == Success) {
        if (actual_type == XA_CARDINAL && nitems == 1)
          pid = *pid_data;

        XFree(pid_data);
      }

      gethostname(hostname, sizeof(hostname));

      if (XGetWindowProperty(display, client_window, XA_WM_CLIENT_MACHINE,
                             0, 256, False, XA_STRING,
                             &actual_type, &actual_format, &nitems, &bytes_after,
                             (unsigned char**)&string_data) == Success) {
        if (actual_type == XA_STRING && nitems > 0) {
          /* If client is on a different host, PID is meaningless. */
          if (strcmp(string_data, hostname) != 0)
            pid = 0;
        }

        XFree(string_data);
      }

      if (pid > 1) {
        printf(" - Client has _NET_WM_PID (%u) and process is local, sending signal KILL,\n", pid);
        kill(pid, SIGKILL);
        return;
      }
    }

    printf(" - Killing client's connection to X.\n");
    XKillClient(display, client_window);
  }
  else /* victim_op == ICONIFY_VICTIM */ {
    printf("xoverkill: Iconifying window %p...\n", (void*)client_window);
    XIconifyWindow(display, client_window, screen_number);
  }
}

void animation_init() {
  XGCValues gc_values;
  int r, g, b, i;
  GC shape_fill_gc;

  frag_area.x = selected_x + h_overlap - 2;
  frag_area.y = selected_y + v_overlap - 2;
  frag_area.width = frag_area.height = 4;
  frag_rate = 1;

  stipple_pixmap = XCreatePixmap(display, overlay_buffer, 16, 16, 1);

  gc_values.graphics_exposures = False;

  gc_values.function = GXclear;
  stipple_gcs[0] = XCreateGC(display, stipple_pixmap,
                             GCGraphicsExposures |
                             GCFunction,
                             &gc_values);
  gc_values.function = GXset;
  stipple_gcs[1] = XCreateGC(display, stipple_pixmap,
                             GCGraphicsExposures |
                             GCFunction,
                             &gc_values);

  gc_values.function = GXcopy;
  gc_values.fill_style = FillStippled;
  gc_values.stipple = stipple_pixmap;
  r = g = b = 0x0000;
  for (i = 0; i < 128; ++i) {
    XColor color;

    if (i < 31) {
      r += 0x0808;
      g += 0x0202;
    }
    else if (i == 31) {
      r = 0xffff;
      g = 0x4040;
    }
    else if (i < 79)
      g += 0x0404;
    else if (i == 79)
      g = 0xffff;
    else if (i >= 83 && i < 104)
      b += 0x0c0c;
    else if (i == 104)
      b = 0xffff;

    color.red = r;
    color.green = g;
    color.blue = b;
    color.pixel = 0;
    XAllocColor(display, screen->cmap, &color);
    gc_values.foreground = color.pixel;

    gradient_gcs[i] = XCreateGC(display, overlay_buffer,
                                GCGraphicsExposures |
                                GCFunction |
                                GCFillStyle |
                                GCStipple |
                                GCForeground,
                                &gc_values);
  }

  gc_values.function = GXcopy;
  gc_values.foreground = BlackPixelOfScreen(screen);
  gc_values.fill_style = FillStippled;
  gc_values.stipple = stipple_pixmap;
  smoke_gc = XCreateGC(display, overlay_buffer,
                       GCGraphicsExposures |
                       GCFunction |
                       GCFillStyle |
                       GCStipple |
                       GCForeground,
                       &gc_values);

  shape_pixmap = XCreatePixmap(display, overlay_buffer,
                               overlay_width, overlay_height, 1);

  gc_values.function = GXset;
  shape_fill_gc = XCreateGC(display, shape_pixmap,
                            GCFunction,
                            &gc_values);
  XFillRectangle(display, shape_pixmap, shape_fill_gc,
                 0, 0, overlay_width, overlay_height);
  XFreeGC(display, shape_fill_gc);

  gc_values.function = GXcopy;
  gc_values.foreground = 0;
  gc_values.fill_style = FillOpaqueStippled;
  gc_values.stipple = stipple_pixmap;
  shape_gcs[SHAPE_GC_CLEAR_STIPPLED] = XCreateGC(display, shape_pixmap,
                                                 GCGraphicsExposures |
                                                 GCFunction |
                                                 GCForeground |
                                                 GCFillStyle |
                                                 GCStipple,
                                                 &gc_values);

  gc_values.function = GXclear;
  shape_gcs[SHAPE_GC_CLEAR_SOLID] = XCreateGC(display, shape_pixmap,
                                              GCGraphicsExposures |
                                              GCFunction,
                                              &gc_values);

  gc_values.function = GXcopy;
  gc_values.foreground = 1;
  gc_values.fill_style = FillOpaqueStippled;
  gc_values.stipple = stipple_pixmap;
  shape_gcs[SHAPE_GC_SET_STIPPLED] = XCreateGC(display, shape_pixmap,
                                               GCGraphicsExposures |
                                               GCFunction |
                                               GCForeground |
                                               GCFillStyle |
                                               GCStipple,
                                               &gc_values);

  max_active_frags = min(overlay_width, overlay_height) / 8;
}

void animation_run() {
  int timeout;
  const int interval = 1000000 / FRAMES_PER_SECOND;

  timeout = interval;
  for (;;) {
    struct timeval then, now;
#   define ELAPSED(THEN, NOW) ((NOW.tv_sec - THEN.tv_sec) * 1000000 + \
                               (NOW.tv_usec - THEN.tv_usec))

    gettimeofday(&then, NULL);

    XSync(display, True);

    if (!animation_advance())
      return;

#ifdef SOUND_ENABLED
    if (sound_ok)
      sound_generate_frame();
#endif

    gettimeofday(&now, NULL);
    timeout -= ELAPSED(then, now);

#ifdef SOUND_ENABLED
    if (timeout < 0)
      sound_generate_frame();
    else
#endif
    if (timeout > 0) {
      usleep(timeout);
      timeout = 0;
    }

    timeout += interval;
  }
}

#define randomize_stipple() {								\
  int stipple_x, stipple_y;								\
  for (stipple_y = 0; stipple_y < 16; ++stipple_y) {					\
    for (stipple_x = 0; stipple_x < 16; ++stipple_x) {					\
      int r = (rand() >> 30) & 1;							\
      XDrawPoint(display, stipple_pixmap, stipple_gcs[r], stipple_x, stipple_y);	\
    }											\
  }											\
}											\

bool animation_advance() {
  bool shaping;
  Frag* f;

  if (n_active_frags < max_active_frags && !no_more_frags)
    frag_create();

  shaping = false;

  for (f = frag_list; f; f = f->next) {
    if (f->stage == FRAG_CLEARING) {
      unsigned r;

      randomize_stipple();

      for (r = MIN_RADIUS; r <= f->current_radius; ++r) {
        XFillArc(display, shape_pixmap, shape_gcs[SHAPE_GC_CLEAR_STIPPLED],
                 f->x + (f->max_radius - r / 2),
                 f->y + (f->max_radius - r / 2),
                 r, r,
                 0, 23040);
      }

      shaping = true;
    }
    else if (f->stage == FRAG_CLEARED) {
      XFillArc(display, shape_pixmap, shape_gcs[SHAPE_GC_CLEAR_SOLID],
               f->x, f->y, f->max_radius * 2, f->max_radius * 2,
               0, 23040);

      shaping = true;
    }
  }

  for (f = frag_list; f; f = f->next) {
    if (f->stage == FRAG_SMOKING) {
      int arc_x, arc_y;
      unsigned arc_diameter;

      randomize_stipple();

      arc_x = f->x + (f->max_radius - f->current_radius);
      arc_y = f->y + (f->max_radius - f->current_radius);
      arc_diameter = f->current_radius * 2;

      XFillArc(display, overlay_buffer, smoke_gc,
               arc_x, arc_y, arc_diameter, arc_diameter,
               0, 23040);

      if (shaping) {
        XFillArc(display, shape_pixmap, shape_gcs[SHAPE_GC_SET_STIPPLED],
                 arc_x, arc_y, arc_diameter, arc_diameter,
                 0, 23040);
      }
    }
  }

  for (f = frag_list; f; f = f->next) {
    if (f->stage == FRAG_EXPANDING) {
      double gradient_offset, gradient_step;
      unsigned r;

      randomize_stipple();

      gradient_offset = 0.;
      gradient_step = 128. / f->current_radius;
      for (r = f->current_radius; r >= MIN_RADIUS; --r) {
        int arc_x, arc_y;
        unsigned arc_diameter;

        arc_x = f->x + (f->max_radius - r);
        arc_y = f->y + (f->max_radius - r);
        arc_diameter = r * 2;

        XFillArc(display, overlay_buffer, gradient_gcs[(int)gradient_offset],
                 arc_x, arc_y, arc_diameter, arc_diameter,
                 0, 23040);
        gradient_offset += gradient_step;

        if (shaping) {
          XFillArc(display, shape_pixmap, shape_gcs[SHAPE_GC_SET_STIPPLED],
                   arc_x, arc_y, arc_diameter, arc_diameter,
                   0, 23040);
        }
      }
    }

    frag_advance(f);
  }

  overlay_blit_area(frag_area.x, frag_area.y, frag_area.width, frag_area.height);

  if (shaping) {
    /* get victim out of the way so it doesn't show through the shape */
    if (!disposed_of_victim) {
      kill_or_iconify_victim();
      disposed_of_victim = true;
    }

    XShapeCombineMask(display, overlay_window, ShapeBounding,
                      0, 0, shape_pixmap, ShapeSet);
  }

  if (frag_area.x > 0) {
    int d = frag_area.x;

    frag_area.x -= frag_rate;
    if (frag_area.x < 0)
      frag_area.x = 0;

    d -= frag_area.x;
    frag_area.width += d;
  }
  if (frag_area.y > 0) {
    int d = frag_area.y;

    frag_area.y -= frag_rate;
    if (frag_area.y < 0)
      frag_area.y = 0;

    d -= frag_area.y;
    frag_area.height += d;
  }
  if (frag_area.x + frag_area.width < overlay_width)
    frag_area.width += frag_rate;
  if (frag_area.y + frag_area.height < overlay_height)
    frag_area.height += frag_rate;

  if (frag_area.x + frag_area.width > overlay_width)
    frag_area.width = overlay_width - frag_area.x;
  if (frag_area.y + frag_area.height > overlay_height)
    frag_area.height = overlay_height - frag_area.y;

  if (!(frag_area.x == 0 && frag_area.y == 0 &&
        frag_area.width == overlay_width && frag_area.height == overlay_height)) {
    ++frag_rate;
  }
  else {
    no_more_frags = true;

    if (n_active_frags == 0)
      return false;
  }
  return true;
}

Frag* frag_create() {
  Frag* f;
  int quadrant_width, quadrant_height;
  unsigned radius;

  f = malloc(sizeof(Frag));

  quadrant_width = frag_area.width / 2;
  quadrant_height = frag_area.height / 2;

  radius = MIN_RADIUS + randn(min(quadrant_width, quadrant_height) / 2 - MIN_RADIUS);
  if (radius > GRADIENT_SIZE)
    radius = GRADIENT_SIZE;

  f->x = frag_area.x + randn(quadrant_width - radius * 2);
  f->y = frag_area.y + randn(quadrant_height - radius * 2);
  if (frag_quadrant == 1 || frag_quadrant == 2)
    f->x += randn(quadrant_width);
  if (frag_quadrant == 2 || frag_quadrant == 3)
    f->y += randn(quadrant_height);
  frag_quadrant = (frag_quadrant + 1) % 4;

  f->current_radius = MIN_RADIUS;
  f->max_radius = radius;

  f->stage = FRAG_EXPANDING;

  f->next = NULL;
  if (!frag_list)
    frag_list = f;
  else {
    Frag* i = frag_list;
    while (i->next) i = i->next;
    i->next = f;
  }

  ++n_active_frags;

  return f;
}

void frag_advance(Frag* f) {
  if (f->current_radius < f->max_radius) {
    switch (f->stage) {
      case FRAG_EXPANDING:
        f->current_radius += MIN_RADIUS;
        break;

      case FRAG_SMOKING:
      case FRAG_CLEARING:
        f->current_radius += MIN_RADIUS * 2;
        break;

      /* not reached */
      case FRAG_CLEARED:
        break;
    }

    if (f->current_radius > f->max_radius)
      f->current_radius = f->max_radius;
  }
  else /* max radius */ {
#ifdef SOUND_ENABLED
    sound_level = 127;
    sound_pulse = SOUND_PULSE_BEGIN;
#endif

    switch (f->stage) {
      case FRAG_EXPANDING: {
        f->stage = FRAG_SMOKING;
        f->current_radius = 2;
      } break;

      case FRAG_SMOKING: {
        f->stage = FRAG_CLEARING;
        f->current_radius = 2;
      } break;

      case FRAG_CLEARING: {
        f->stage = FRAG_CLEARED;

        --n_active_frags;
      } break;

      case FRAG_CLEARED: {
        /* do nothing */
      } break;
    }
  }
}

void frag_destroy(Frag* f) {
  if (f == frag_list)
    frag_list = f->next;
  else {
    Frag* i = frag_list;
    while (i->next != f) i = i->next;
    i->next = f->next;
  }

  free(f);
}

#ifdef SOUND_ENABLED

void sound_init() {
  sound_fd = open(sound_device, O_WRONLY|O_NONBLOCK);
  sound_ok = (sound_fd >= 0);
}

void sound_generate_frame() {
  signed char data[8192 / FRAMES_PER_SECOND * 2];
  int i;
  int v = 42, sign = 1;

  for (i = 0; i < sizeof(data); ++i) {
    if (i % 8 == 0) {
      v = sign * (sound_pulse == SOUND_PULSE_BEGIN ? sound_level : randn(sound_level));
      if (v < -128)
        v = -128;
      else if (v > 127)
        v = 127;

      if (sound_pulse && --sound_pulse == 0) {
        sound_level >>= 1;
        if (!sound_level)
          sound_level = 1;
      }
    }

    data[i] = (signed char)v;

    if ((rand() >> 30) & 1)
      sign = -sign;
  }

  write(sound_fd, data, sizeof(data));
}
#endif
